from collections import Iterable
from itertools import zip_longest
from typing import Tuple, List
from typing import Pattern as RegexPattern

from pampy.helpers import *

_ = ANY = UnderscoreType()
HEAD = HeadType()
REST = TAIL = TailType()


def run(action, var):
    if callable(action):
        if isinstance(var, Iterable):
            try:
                return action(*var)
            except TypeError as err:
                raise MatchError(get_lambda_args_error_msg(action, var, err))
        elif isinstance(var, BoxedArgs):
            return action(var.get())
        else:
            return action(var)
    else:
        return action


def match_value(pattern, value) -> Tuple[bool, List]:
    if value is PaddedValue:
        return False, []
    elif isinstance(pattern, (int, float, str, bool)):
        eq = pattern == value
        type_eq = type(pattern) == type(value)
        return eq and type_eq, []
    elif pattern is None:
        return value is None, []
    elif isinstance(pattern, type):
        if isinstance(value, pattern):
            return True, [value]
        else:
            return False, []
    elif isinstance(pattern, (list, tuple)):
        return match_iterable(pattern, value)
    elif isinstance(pattern, dict):
        return match_dict(pattern, value)
    elif callable(pattern):
        return_value = pattern(value)
        if return_value is True:
            return True, [value]
        elif return_value is False:
            pass
        else:
            raise MatchError("Warning! pattern function %s is not returning a boolean, but instead %s" %
                             (pattern, return_value))
    elif isinstance(pattern, RegexPattern):
        rematch = pattern.search(value)
        if rematch is not None:
            return True, list(rematch.groups())
    elif pattern is _:
        return True, [value]
    elif pattern is HEAD or pattern is TAIL:
        raise MatchError("HEAD or TAIL should only be used inside an Iterable (list or tuple).")
    elif is_dataclass(pattern) and pattern.__class__ == value.__class__:
        return match_dict(pattern.__dict__, value.__dict__)
    return False, []


def match_dict(pattern, value) -> Tuple[bool, List]:
    if not isinstance(value, dict) or not isinstance(pattern, dict):
        return False, []

    total_extracted = []
    still_usable_value_keys = set(value.keys())
    still_usable_pattern_keys = set(pattern.keys())
    for pkey, pval in pattern.items():
        if pkey not in still_usable_pattern_keys:
            continue
        matched_left_and_right = False
        for vkey, vval in value.items():
            if vkey not in still_usable_value_keys:
                continue
            if pkey not in still_usable_pattern_keys:
                continue
            key_matched, key_extracted = match_value(pkey, vkey)
            if key_matched:
                value_matched, value_extracted = match_value(pval, vval)
                if value_matched:
                    total_extracted += key_extracted + value_extracted
                    matched_left_and_right = True
                    still_usable_pattern_keys.remove(pkey)
                    still_usable_value_keys.remove(vkey)
                    break
        if not matched_left_and_right:
            return False, []
    return True, total_extracted


def only_padded_values_follow(padded_pairs, i):
    i += 1
    while i < len(padded_pairs):
        pattern, value = padded_pairs[i]
        if pattern is not PaddedValue:
            return False
        i += 1
    return True


def match_iterable(patterns, values) -> Tuple[bool, List]:
    if not isinstance(patterns, Iterable) or not isinstance(values, Iterable):
        return False, []

    total_extracted = []
    padded_pairs = list(zip_longest(patterns, values, fillvalue=PaddedValue))

    for i, (pattern, value) in enumerate(padded_pairs):
        if pattern is HEAD:
            if i != 0:
                raise MatchError("HEAD can only be in first position of a pattern.")
            else:
                if value is PaddedValue:
                    return False, []
                else:
                    total_extracted += [value]
        elif pattern is TAIL:
            if not only_padded_values_follow(padded_pairs, i):
                raise MatchError("TAIL must me in last position of the pattern.")
            else:
                tail = [value for (pattern, value) in padded_pairs[i:] if value is not PaddedValue]
                total_extracted.append(tail)
                break
        else:
            matched, extracted = match_value(pattern, value)
            if not matched:
                return False, []
            else:
                total_extracted += extracted
    return True, total_extracted





def match(var, *args, default=NoDefault, strict=True):
    """
    Match `var` against a number of potential patterns.

    Example usage:
    ```
    match(x,
        3,              "this matches the number 3",
        int,            "matches any integer",
        (str, int),     lambda a, b: "a tuple (a, b) you can use in a function",
        [1, 2, _],      "any list of 3 elements that begins with [1, 2]",
        {'x': _},       "any dict with a key 'x' and any value associated",
        _,              "anything else"
    )
    ```

    :param var: The variable to test patterns against.
    :param args: Alternating patterns and actions. There must be an action for every pattern specified.
                 Patterns can take many forms, see README.md for examples.
                 Actions can be either a literal value or a callable which will be called with the arguments that were
                    matched in corresponding pattern.
    :param default: If `default` is specified then it will be returned if none of the patterns match.
                    If `default` is unspecified then a `MatchError` will be thrown instead.
    :return: The result of the action which corresponds to the first matching pattern.
    """
    if len(args) % 2 != 0:
        raise MatchError("Every guard must have an action.")

    if default is NoDefault and strict is False:
        default = False

    pairs = list(pairwise(args))
    patterns = [patt for (patt, action) in pairs]

    for patt, action in pairs:
        matched_as_value, args = match_value(patt, var)

        if matched_as_value:
            lambda_args = args if len(args) > 0 else BoxedArgs(var)
            return run(action, lambda_args)

    if default is NoDefault:
        if _ not in patterns:
            raise MatchError("'_' not provided. This case is not handled:\n%s" % str(var))
    else:
        return default


class MatchError(Exception):
    def __init__(self, msg):
        super().__init__(msg)



import inspect


class UnderscoreType:
    def __repr__(self):
        return '_'


class HeadType:
    def __repr__(self):
        return 'HEAD'


class TailType:
    def __repr__(self):
        return 'TAIL'


class PaddedValueType:
    def __repr__(self):
        return 'PaddedValue'


class NoDefaultType:
    pass


NoDefault = NoDefaultType()

PaddedValue = PaddedValueType()


class BoxedArgs:
    def __init__(self, obj):
        self.obj = obj

    def get(self):
        return self.obj


def pairwise(l):
    i = 0
    while i < len(l):
        yield l[i], l[i + 1]
        i += 2


def get_lambda_args_error_msg(action, var, err):
    try:
        code = inspect.getsource(action)
        return "Error passing arguments %s here:\n%s\n%s" % (var, code, err)
    except OSError:
        return "Error passing arguments %s:\n%s" % (var, err)


def is_dataclass(value):
    """
    Dataclass support is only enabled in Python 3.7+, or in 3.6 with the `dataclasses` backport installed
    """
    try:
        from dataclasses import is_dataclass
        return is_dataclass(value)
    except ImportError:
        return False



__version__ = '0.2.1'

import sys

major, minor = sys.version_info.major, sys.version_info.minor

if major < 3:
    sys.exit("Sorry, Python 2 is not supported. You need Python >= 3.6 for Pampy.")
elif minor < 6:
    sys.exit("Sorry, You need Python >= 3.6 for Pampy.")


from pampy.pampy import match, _, ANY, HEAD, TAIL, REST, MatchError
from pampy.pampy import match_value, match_iterable, match_dict




import unittest

from functools import reduce

from pampy import match, REST, TAIL, HEAD, _, match_value, match_iterable


class PampyElaborateTests(unittest.TestCase):

    def test_fibonacci(self):
        def fib(n):
            return match(n,
                         1, 1,
                         2, 1,
                         _, lambda x: fib(x - 1) + fib(x - 2))

        self.assertEqual(fib(1), 1)
        self.assertEqual(fib(7), 13)

    def test_slide1(self):
        _input = [1, 2, 3]
        pattern = [1, _, 3]
        action = lambda x: "it's {}".format(x)
        self.assertEqual(match(_input, pattern, action), "it's 2")

    def test_parser(self):
        def parser(exp):
            return match(exp,
                         3, "the integer 3",
                         float, "any float number",
                         int, "any integer",
                         "ciao", "the string ciao",
                         dict, "any dictionary",
                         str, "any string",
                         (int, int), "a tuple made of two ints",
                         [1], "the list [1]",
                         [1, 2, 3], "the list [1, 2, 3]",
                         [1, _, 3], "the list [1, _, 3]",
                         (str, str), lambda a, b: "%s %s" % (a, b),
                         [1, 2, _], lambda x: "the list [1, 2, _]",
                         [1, 2, 4], "the list [1, 2, 4]",  # this can never be matched

                         [1, [2, _], _], lambda a, b: "[1, [2, %s], %s]" % (a, b),
                         )

        assert parser(3) == "the integer 3"
        assert parser(5) == "any integer"
        assert parser("ciao") == "the string ciao"
        assert parser("x") == "any string"
        assert parser({'a': 1}) == "any dictionary"
        assert parser([1]) == "the list [1]"
        assert parser([1, 2, 3]) == "the list [1, 2, 3]"
        assert parser([1, 2, 4]) == "the list [1, 2, _]"
        assert parser([1, 3, 3]) == "the list [1, _, 3]"
        assert parser(("hello", "world")) == "hello world"
        assert parser([1, [2, 3], 4]) == "[1, [2, 3], 4]"

    def test_lisp(self):
        # A Lisp interpreter in 5 lines
        def lisp(exp):
            return match(exp,
                         int, lambda x: x,
                         callable, lambda x: x,
                         (callable, REST), lambda f, rest: f(*map(lisp, rest)),
                         tuple, lambda t: list(map(lisp, t)),
                         )

        plus = lambda a, b: a + b
        minus = lambda a, b: a - b
        self.assertEqual(lisp((plus, 1, 2)), 3)
        self.assertEqual(lisp((plus, 1, (minus, 4, 2))), 3)
        self.assertEqual(lisp((reduce, plus, (1, 2, 3))), 6)
        self.assertEqual(lisp((reduce, plus, (range, 10))), 45)

    def test_myzip(self):
        def myzip(a, b):
            return match((a, b),
                         ([], []), [],
                         ([_, TAIL], [_, TAIL]), lambda ha, ta, hb, tb: [(ha, hb)] + myzip(ta, tb)
                         )

        self.assertEqual(myzip([1, 2, 3], [4, 5, 6]), [(1, 4), (2, 5), (3, 6)])
        self.assertEqual(myzip(range(5), range(5)), list(zip(range(5), range(5))))

    def test_lambda_cond(self):
        cond = lambda x: x < 10
        self.assertEqual(match(3, cond, "action", _, "else"), "action")
        self.assertEqual(match(11, cond, "action1", _, "else"), "else")

    def test_lambda_cond_arg_passing(self):
        def f(x):
            return match(x,
                         lambda x: x % 2 == 0, lambda x: "even %d" % x,
                         lambda x: x % 2 != 0, lambda x: "odd %d" % x
                         )

        self.assertEqual(f(3), "odd 3")
        self.assertEqual(f(18), "even 18")
        self.assertEqual(f(18), "even 18")

    def test_animals(self):
        pets = [
            {'type': 'dog', 'pet-details': {'name': 'carl', 'cuteness': 4}},
            {'type': 'dog', 'pet-details': {'name': 'john', 'cuteness': 3}},
            {'type': 'cat', 'pet-details': {'name': 'fuffy', 'cuty': 4.6}},
            {'type': 'cat', 'cat-details': {'name': 'bonney', 'cuty': 7}},
        ]

        def avg_cuteness_pampy():
            cutenesses = []
            for pet in pets:
                match(pet,
                      {_: {"cuteness": _}}, lambda key, x: cutenesses.append(x),
                      {_: {"cuty": _}}, lambda key, x: cutenesses.append(x)
                      )
            return sum(cutenesses) / len(cutenesses)

        self.assertEqual(avg_cuteness_pampy(), (4 + 3 + 4.6 + 7) / 4)



import setuptools

from pampy import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pampy",
    version=__version__,
    author="Claudio Santini",
    author_email="hireclaudio@gmail.com",
    description="The Pattern Matching for Python you always dreamed of",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/santinic/pampy",
    packages=setuptools.find_packages(),
    platforms='any',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>3.6'
)
